#include <stdlib.h>
#include <assert.h>
#include "list.h"
#include "list_elm.h"

list_t * new_list() {
    list_t * L = calloc(1, sizeof(list_t));
    assert(L);
    L->head = NULL;
    L->tail = NULL;
    L->numelm = 0;
    return L;
}

void del_list(list_t ** L, void (*ptr_del)(void *)) {
    assert(L && *L);
    list_t * list = *L;
    list_elm_t * E = list->head;
    while (E) {
        list_elm_t * temp = E;
        E = E->suc;
        if(ptr_del) {
            ptr_del(temp->data);
        }
        free(temp);
    }
    free(*L);
    *L = NULL;
}

bool empty_list(list_t * L) {
    assert(L);
    return L->numelm == 0;
}

list_elm_t * get_head(list_t * L) {
    assert(L);
    return L->head;
}

list_elm_t * get_tail(list_t * L) {
    assert(L);
    return L->tail;
}

void view_list(list_t * L, void (*ptr_view)(void *)) {
    assert(L && ptr_view);
    for (list_elm_t * E = L->head; E != NULL; E = get_suc(E)) {
        ptr_view(E->data);
    }
}

void cons(list_t * L, void * data) {
    list_elm_t * E = new_list_elm(data);
    E->suc = L->head;
    if (L->head != NULL) {
        L->head->pred = E;
    }
    L->head = E;
    if (L->tail == NULL) {
        L->tail = E;
    }
    L->numelm++;
}

void queue(list_t * L, void * data) {
    list_elm_t * E = new_list_elm(data);
    if(empty_list(L)) {
        L->head = E;
        L->tail = E;
    } else {
        L->tail->suc = E;
        E->pred = L->tail;
        L->tail = E;
    }
    L->numelm++;
}

void insert_after(list_t * L, void * data, list_elm_t * elm) {
    if(elm == NULL) return;
    list_elm_t * E = new_list_elm(data);
    E->suc = elm->suc;
    E->pred = elm;
    if(elm->suc != NULL) {
        elm->suc->pred = E;
    } else {
        L->tail = E;
    }
    elm->suc = E;
    L->numelm++;
}

void ordered_insert(list_t * L, void * data, int (*cmpFct)(void*, void*)) {
    if(empty_list(L) || cmpFct(L->head->data, data) > 0) {
        cons(L, data);
    } else {
        list_elm_t * current = L->head;
        while(get_suc(current) != NULL && cmpFct(get_suc(current)->data, data) <= 0) {
            current = get_suc(current);
        }
        insert_after(L, data, current);
    }
}
